@@ -4,7 +4,7 @@
 import os
 import sys
 
-from pyfygentlescrap.warnings import (  # noqa F401
+from pyfygentlescrap.warnings import (
     InvalidParameterWarning,
     InvalidRegionWarning,
     InvalidResponseWarning,
@@ -21,7 +21,7 @@ from pyfygentlescrap.yahoo.yahoo_shared import yahoo_session
 from pyfygentlescrap.yahoo.yahoo_ticker import yahoo_ticker
 
 
-__version__ = "0.4.0"
+__version__ = "0.4.11"
 
 
 sys.path.insert(0, os.path.abspath(".."))
